@@ -1,7 +1,27 @@
 #include <libplatform/libplatform.h>
 #include "V8_types.h"
 
-#if (V8_MAJOR_VERSION * 100 + V8_MINOR_VERSION) < 803
+/* use conditional apis below */
+#define V8_VERSION_TOTAL (V8_MAJOR_VERSION * 100 + V8_MINOR_VERSION)
+
+/* we dont assume <node/node_version.h> is installed */
+#ifdef ISNODEJS
+#if V8_MAJOR_VERSION == 10
+#define NODEJS_LTS_API 18
+#elif V8_MAJOR_VERSION == 9
+#define NODEJS_LTS_API 16
+#elif V8_MAJOR_VERSION < 9
+#define NODEJS_LTS_API 14
+#endif
+#endif
+
+#if V8_VERSION_TOTAL < 908 || NODEJS_LTS_API == 16
+#define FixedArrayParam
+#else
+#define FixedArrayParam ,v8::Local<v8::FixedArray> import_arributes
+#endif
+
+#if V8_VERSION_TOTAL < 803
 #define PerformMicrotaskCheckpoint RunMicrotasks
 #endif
 
@@ -24,7 +44,7 @@ v8::Local<T> safe_to_local(v8::MaybeLocal<T> x){
   return x.IsEmpty() ? v8::Local<T>() : x.ToLocalChecked();
 }
 
-void ctx_finalizer( v8::Persistent<v8::Context>* context ){
+void ctx_finalizer(ctx_type* context ){
   if(context)
     context->Reset();
   delete context;
@@ -33,6 +53,17 @@ void ctx_finalizer( v8::Persistent<v8::Context>* context ){
 static v8::Isolate* isolate = NULL;
 static v8::Platform* platformptr = NULL;
 
+static std::string read_text(std::string filename) {
+  if(filename.empty() || (filename.at(0) != '.' && filename.at(0) != '/'))
+    throw std::runtime_error("Invalid module: " + filename + " (paths should begin with . or /)");
+  std::ifstream t(filename);
+  if(t.fail())
+    throw std::runtime_error("Failed to open file: " + filename);
+  std::stringstream buffer;
+  buffer << t.rdbuf();
+  return buffer.str();
+}
+
 // Extracts a C string from a V8 Utf8Value.
 static const char* ToCString(const v8::String::Utf8Value& value) {
   return *value ? *value : "<string conversion failed>";
@@ -52,6 +83,98 @@ static void fatal_cb(const char* location, const char* message){
   REprintf("V8 FATAL ERROR in %s: %s", location, message);
 }
 
+static v8::Local<v8::Module> read_module(std::string filename, v8::Local<v8::Context> context);
+
+static v8::MaybeLocal<v8::Module> ResolveModuleCallback(v8::Local<v8::Context> context, v8::Local<v8::String> specifier
+                                                        FixedArrayParam, v8::Local<v8::Module> referrer) {
+  v8::String::Utf8Value name(context->GetIsolate(), specifier);
+  try {
+    return read_module(*name, context);
+  }
+  catch(const std::exception& err) {
+    isolate->ThrowException(ToJSString(err.what()));
+  }
+  return v8::Local<v8::Module>();
+};
+
+static v8::MaybeLocal<v8::Promise> dynamic_module_loader(v8::Local<v8::Context> context, v8::Local<v8::String> specifier) {
+  v8::Local<v8::Promise::Resolver> resolver = v8::Promise::Resolver::New(context).ToLocalChecked();
+  v8::MaybeLocal<v8::Promise> promise(resolver->GetPromise());
+  v8::String::Utf8Value name(context->GetIsolate(), specifier);
+  try {
+    v8::Local<v8::Module> module = read_module(*name, context);
+    resolver->Resolve(context, module->GetModuleNamespace()).FromMaybe(false);
+  } catch(const std::exception& err) {
+    resolver->Reject(context, ToJSString(err.what())).FromMaybe(false);
+  } catch(...) {
+    resolver->Reject(context, ToJSString("Unknown failure loading dynamic module")).FromMaybe(false);
+  }
+  return promise;
+}
+
+static v8::MaybeLocal<v8::Promise> ResolveDynamicModuleCallback(
+    v8::Local<v8::Context> context,
+#if V8_VERSION_TOTAL >= 908
+    v8::Local<v8::Data> host_defined_options,
+    v8::Local<v8::Value> resource_name,
+#elif V8_VERSION_TOTAL >= 603
+    v8::Local<v8::ScriptOrModule> referrer,
+#else
+    v8::Local<v8::String> referrer,
+#endif
+    v8::Local<v8::String> specifier
+    FixedArrayParam
+) {
+  return dynamic_module_loader(context, specifier);
+}
+
+static v8::ScriptOrigin make_origin(std::string filename){
+#if V8_VERSION_TOTAL < 908 || NODEJS_LTS_API == 16
+  return v8::ScriptOrigin(ToJSString( filename.c_str()), v8::Integer::New(isolate, 0),
+                          v8::Integer::New(isolate, 0), v8::False(isolate), v8::Local<v8::Integer>(),
+                          v8::Local<v8::Value>(), v8::False(isolate), v8::False(isolate), v8::True(isolate));
+#elif V8_VERSION_TOTAL < 1201
+  return v8::ScriptOrigin(isolate,ToJSString( filename.c_str()), 0, 0, false, -1,
+                          v8::Local<v8::Value>(), false, false, true);
+#else
+  return v8::ScriptOrigin(ToJSString( filename.c_str()), 0, 0, false, -1,
+                          v8::Local<v8::Value>(), false, false, true);
+#endif
+}
+
+static std::string throw_js_err(v8::Local<v8::Value> Exception, std::string filename){
+  std::string errmsg(std::string("Failed to import ES module '") + filename + "': " + *v8::String::Utf8Value(isolate, Exception));
+  throw std::runtime_error(errmsg);
+}
+
+/* Helper fun that compiles JavaScript source code */
+static v8::Local<v8::Module> read_module(std::string filename, v8::Local<v8::Context> context){
+  v8::Local<v8::String> source_text = ToJSString(read_text(filename).c_str());
+  if(source_text.IsEmpty())
+    throw std::runtime_error("Failed to read module file (check memory/stack limits.");
+  v8::TryCatch trycatch(isolate);
+  v8::ScriptCompiler::Source source(source_text, make_origin(filename));
+  v8::Local<v8::Module> module;
+  if (!v8::ScriptCompiler::CompileModule(isolate, &source).ToLocal(&module)){
+    if(trycatch.HasCaught())
+      throw_js_err(trycatch.Exception(), filename);
+    throw std::runtime_error("Failed to run CompileModule() source.");
+  }
+  if(!module->InstantiateModule(context, ResolveModuleCallback).FromMaybe(false)){
+    if(trycatch.HasCaught())
+      throw_js_err(trycatch.Exception(), filename);
+    throw std::runtime_error("Failed to run InstantiateModule().");
+  }
+  v8::Local<v8::Value> retValue;
+  if (!module->Evaluate(context).ToLocal(&retValue)){
+    if(trycatch.HasCaught())
+      throw_js_err(trycatch.Exception(), filename);
+    throw std::runtime_error("Failure loading module");
+  }
+  return module;
+}
+
+
 // [[Rcpp::init]]
 void start_v8_isolate(void *dll){
 #ifdef V8_ICU_DATA_PATH
@@ -60,7 +183,7 @@ void start_v8_isolate(void *dll){
     v8::V8::InitializeICUDefaultLocation(V8_ICU_DATA_PATH);
   }
 #endif
-#if (V8_MAJOR_VERSION * 100 + V8_MINOR_VERSION) >= 704
+#if V8_VERSION_TOTAL >= 704
   std::unique_ptr<v8::Platform> platform = v8::platform::NewDefaultPlatform();
   v8::V8::InitializePlatform(platform.get());
   platformptr = platform.get();
@@ -89,6 +212,7 @@ void start_v8_isolate(void *dll){
   uintptr_t CurrentStackPosition = reinterpret_cast<uintptr_t>(__builtin_frame_address(0));
   isolate->SetStackLimit(CurrentStackPosition - kWorkerMaxStackSize);
 #endif
+  isolate->SetHostImportModuleDynamicallyCallback(ResolveDynamicModuleCallback);
 }
 
 /* Helper fun that compiles JavaScript source code */
@@ -110,7 +234,7 @@ static void pump_promises(){
 /* Try to resolve pending promises */
 static void ConsolePump(const v8::FunctionCallbackInfo<v8::Value>& args) {
   pump_promises();
-  args.GetReturnValue().Set(v8::Undefined(args.GetIsolate()));
+  //args.GetReturnValue().Set(v8::Undefined(args.GetIsolate()));
 }
 
 
@@ -122,7 +246,7 @@ static void ConsoleLog(const v8::FunctionCallbackInfo<v8::Value>& args) {
     Rprintf("%s", ToCString(str));
   }
   Rprintf("\n");
-  args.GetReturnValue().Set(v8::Undefined(args.GetIsolate()));
+  //args.GetReturnValue().Set(v8::Undefined(args.GetIsolate()));
 }
 
 /* console.warn */
@@ -130,10 +254,9 @@ static void ConsoleWarn(const v8::FunctionCallbackInfo<v8::Value>& args) {
   for (int i=0; i < args.Length(); i++) {
     v8::HandleScope handle_scope(args.GetIsolate());
     v8::String::Utf8Value str(args.GetIsolate(), args[i]);
-    Rf_warningcall_immediate(R_NilValue, ToCString(str));
+    Rf_warningcall_immediate(R_NilValue, "%s", ToCString(str));
   }
-  Rprintf("\n");
-  args.GetReturnValue().Set(v8::Undefined(args.GetIsolate()));
+  //args.GetReturnValue().Set(v8::Undefined(args.GetIsolate()));
 }
 
 /* console.error */
@@ -141,7 +264,7 @@ static void ConsoleError(const v8::FunctionCallbackInfo<v8::Value>& args) {
   if(args.Length()){
     args.GetIsolate()->ThrowException(args[0]);
   }
-  args.GetReturnValue().Set(v8::Undefined(args.GetIsolate()));
+  //args.GetReturnValue().Set(v8::Undefined(args.GetIsolate()));
 }
 
 void r_callback(std::string cb, const v8::FunctionCallbackInfo<v8::Value>& args) {
@@ -209,10 +332,13 @@ static Rcpp::RObject convert_object(v8::Local<v8::Value> value){
     v8::Local<v8::ArrayBuffer> buffer = value->IsArrayBufferView() ?
     value.As<v8::ArrayBufferView>()->Buffer() : value.As<v8::ArrayBuffer>();
     Rcpp::RawVector data(buffer->ByteLength());
-#if (V8_MAJOR_VERSION * 100 + V8_MINOR_VERSION) >= 901
-    memcpy(data.begin(), buffer->GetBackingStore()->Data(), data.size());
-#else
+#if V8_VERSION_TOTAL >= 1005 || NODEJS_LTS_API == 18
+    memcpy(data.begin(), buffer->Data(), data.size());
+#elif V8_VERSION_TOTAL < 901 || NODEJS_LTS_API == 16
     memcpy(data.begin(), buffer->GetContents().Data(), data.size());
+#else
+    /* Try to avoid this API: github.com/jeroen/V8/issues/152 */
+    memcpy(data.begin(), buffer->GetBackingStore()->Data(), data.size());
 #endif
     return data;
   } else {
@@ -225,7 +351,7 @@ static Rcpp::RObject convert_object(v8::Local<v8::Value> value){
 }
 
 // [[Rcpp::export]]
-Rcpp::RObject context_eval(Rcpp::String src, Rcpp::XPtr< v8::Persistent<v8::Context> > ctx, bool serialize = false, bool await = false){
+Rcpp::RObject context_eval(Rcpp::String src, ctxptr ctx, bool serialize = false, bool await = false){
   // Test if context still exists
   if(!ctx)
     throw std::runtime_error("v8::Context has been disposed.");
@@ -236,11 +362,12 @@ Rcpp::RObject context_eval(Rcpp::String src, Rcpp::XPtr< v8::Persistent<v8::Cont
   // Create a scope
   v8::Isolate::Scope isolate_scope(isolate);
   v8::HandleScope handle_scope(isolate);
-  v8::Context::Scope context_scope(ctx.checked_get()->Get(isolate));
+  v8::Local<v8::Context> context = ctx.checked_get()->Get(isolate);
+  v8::Context::Scope context_scope(context);
 
   // Compile source code
   v8::TryCatch trycatch(isolate);
-  v8::Local<v8::Script> script = compile_source(src, ctx.checked_get()->Get(isolate));
+  v8::Local<v8::Script> script = compile_source(src, context);
   if(script.IsEmpty()) {
     v8::String::Utf8Value exception(isolate, trycatch.Exception());
     if(*exception){
@@ -251,7 +378,7 @@ Rcpp::RObject context_eval(Rcpp::String src, Rcpp::XPtr< v8::Persistent<v8::Cont
   }
 
   // Run the script to get the result.
-  v8::MaybeLocal<v8::Value> res = script->Run(ctx.checked_get()->Get(isolate));
+  v8::MaybeLocal<v8::Value> res = script->Run(context);
   v8::Local<v8::Value> result = safe_to_local(res);
   if(result.IsEmpty()){
     v8::String::Utf8Value exception(isolate, trycatch.Exception());
@@ -296,7 +423,7 @@ Rcpp::RObject context_eval(Rcpp::String src, Rcpp::XPtr< v8::Persistent<v8::Cont
 }
 
 // [[Rcpp::export]]
-bool write_array_buffer(Rcpp::String key, Rcpp::RawVector data, Rcpp::XPtr< v8::Persistent<v8::Context> > ctx){
+bool write_array_buffer(Rcpp::String key, Rcpp::RawVector data, ctxptr ctx){
   // Test if context still exists
   if(!ctx)
     throw std::runtime_error("v8::Context has been disposed.");
@@ -311,10 +438,13 @@ bool write_array_buffer(Rcpp::String key, Rcpp::RawVector data, Rcpp::XPtr< v8::
   // Initiate ArrayBuffer and ArrayBufferView (uint8 typed array)
   v8::Local<v8::ArrayBuffer> buffer = v8::ArrayBuffer::New(isolate, data.size());
   v8::Local<v8::Uint8Array> typed_array = v8::Uint8Array::New(buffer, 0, data.size());
-#if (V8_MAJOR_VERSION * 100 + V8_MINOR_VERSION) >= 901
-  memcpy(buffer->GetBackingStore()->Data(), data.begin(), data.size());
-#else
+
+#if V8_VERSION_TOTAL >= 1005 || NODEJS_LTS_API == 18
+  memcpy(buffer->Data(), data.begin(), data.size());
+#elif V8_VERSION_TOTAL < 901 || NODEJS_LTS_API == 16
   memcpy(buffer->GetContents().Data(), data.begin(), data.size());
+#else
+  memcpy(buffer->GetBackingStore()->Data(), data.begin(), data.size());
 #endif
 
   // Assign to object (delete first if exists)
@@ -326,7 +456,7 @@ bool write_array_buffer(Rcpp::String key, Rcpp::RawVector data, Rcpp::XPtr< v8::
 }
 
 // [[Rcpp::export]]
-bool context_validate(Rcpp::String src, Rcpp::XPtr< v8::Persistent<v8::Context> > ctx) {
+bool context_validate(Rcpp::String src, ctxptr ctx) {
 
   // Test if context still exists
   if(!ctx)
@@ -347,7 +477,7 @@ bool context_validate(Rcpp::String src, Rcpp::XPtr< v8::Persistent<v8::Context>
 }
 
 // [[Rcpp::export]]
-bool context_null(Rcpp::XPtr< v8::Persistent<v8::Context>> ctx) {
+bool context_null(ctxptr ctx) {
   // Test if context still exists
   return(!ctx);
 }
@@ -393,11 +523,6 @@ ctxptr make_context(bool set_console){
     if(context->Global()->Set(context, console, console_template()).IsNothing())
       Rcpp::warning("Could not set console.");
   }
-  v8::Persistent<v8::Context> *ptr = new v8::Persistent<v8::Context>(isolate, context);
+  ctx_type *ptr = new ctx_type(isolate, context);
   return ctxptr(ptr);
 }
-
-// [[Rcpp::export]]
-bool context_enable_typed_arrays( Rcpp::XPtr< v8::Persistent<v8::Context> > ctx ){
-  return true;
-}
